@@ -1,4 +1,3 @@
-from .asyncio import *
 from .curio import *
 from .framing import *
 from .jsonrpc import *
@@ -6,12 +5,12 @@ from .socks import *
 from .session import *
 from .util import *
 
-_version_str = '0.11.0'
+
+_version_str = '0.13.0'
 _version = tuple(int(part) for part in _version_str.split('.'))
 
 
-__all__ = (asyncio.__all__ +
-           curio.__all__ +
+__all__ = (curio.__all__ +
            framing.__all__ +
            jsonrpc.__all__ +
            socks.__all__ +
