@@ -36,25 +36,33 @@
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 # OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
+from asyncio import (
+    CancelledError, get_event_loop, Queue, Event, Lock, Semaphore, sleep
+)
 from collections import deque
 from contextlib import suppress
 from functools import partial
 import logging
 import sys
 
-from aiorpcx.asyncio import (
-    CancelledError, get_event_loop, Queue, Event, Lock, Semaphore, sleep, current_task
-)
-from aiorpcx.util import normalize_corofunc, check_task
+from aiorpcx.util import instantiate_coroutine, check_task
 
 
 __all__ = (
+    'Queue', 'Event', 'Lock', 'Semaphore', 'sleep', 'CancelledError',
     'run_in_thread', 'spawn', 'spawn_sync', 'TaskGroup',
     'TaskTimeout', 'TimeoutCancellationError', 'UncaughtTimeoutError',
     'timeout_after', 'timeout_at', 'ignore_after', 'ignore_at',
 )
 
 
+if sys.version_info >= (3, 7):
+    from asyncio import current_task
+else:
+    from asyncio import Task
+    current_task = Task.current_task
+
+
 async def run_in_thread(func, *args):
     '''Run a function in a separate thread, and await its completion.'''
     return await get_event_loop().run_in_executor(None, func, *args)
@@ -65,7 +73,7 @@ async def spawn(coro, *args, loop=None, report_crash=True):
 
 
 def spawn_sync(coro, *args, loop=None, report_crash=True):
-    coro = normalize_corofunc(coro, args)
+    coro = instantiate_coroutine(coro, args)
     loop = loop or get_event_loop()
     task = loop.create_task(coro)
     if report_crash:
@@ -321,7 +329,7 @@ class TimeoutAfter(object):
 
 
 async def _timeout_after_func(seconds, absolute, coro, args):
-    coro = normalize_corofunc(coro, args)
+    coro = instantiate_coroutine(coro, args)
     async with TimeoutAfter(seconds, absolute=absolute):
         return await coro
 
@@ -371,7 +379,7 @@ def timeout_at(clock, coro=None, *args):
 
 
 async def _ignore_after_func(seconds, absolute, coro, args, timeout_result):
-    coro = normalize_corofunc(coro, args)
+    coro = instantiate_coroutine(coro, args)
     async with TimeoutAfter(seconds, absolute=absolute, ignore=True):
         return await coro
 
