@@ -24,17 +24,18 @@
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 
-__all__ = ('Connector', 'RPCSession', 'MessageSession', 'Server',
+__all__ = ('Connector', 'RPCSession', 'MessageSession', 'Server', 'ExcessiveSessionCostError',
            'BatchError')
 
 
 import asyncio
 import logging
+from math import ceil
 import time
 
 from aiorpcx.curio import (
     Event, TaskGroup, TaskTimeout, CancelledError,
-    spawn, timeout_after, spawn_sync, ignore_after
+    timeout_after, spawn_sync, ignore_after, sleep
 )
 from aiorpcx.framing import (
     NewlineFramer, BitcoinFramer,
@@ -44,7 +45,6 @@ from aiorpcx.jsonrpc import (
     Request, Batch, Notification, ProtocolError, RPCError, FinalRPCError,
     JSONRPC, JSONRPCv2, JSONRPCConnection
 )
-from aiorpcx.util import Concurrency
 
 
 class Connector(object):
@@ -67,13 +67,50 @@ class Connector(object):
     async def __aenter__(self):
         _transport, self.protocol = await self.create_connection()
         # By default, do not limit outgoing connections
-        self.protocol.bw_limit = 0
+        self.ru_hard_limit = 0
         return self.protocol
 
     async def __aexit__(self, exc_type, exc_value, traceback):
         await self.protocol.close()
 
 
+class ExcessiveSessionCostError(RuntimeError):
+    pass
+
+
+class Concurrency(object):
+
+    def __init__(self, target):
+        self._target = int(target)
+        self._semaphore = asyncio.Semaphore(self._target)
+        self._sem_value = self._target
+
+    async def _retarget_semaphore(self):
+        while self._sem_value != self._target:
+            if self._target <= 0:
+                raise ExcessiveSessionCostError
+            if self._sem_value > self._target:
+                await self._semaphore.acquire()
+                self._sem_value -= 1
+            else:
+                self._semaphore.release()
+                self._sem_value += 1
+
+    @property
+    def max_concurrent(self):
+        return self._target
+
+    def set_target(self, target):
+        self._target = int(target)
+
+    async def __aenter__(self):
+        await self._retarget_semaphore()
+        await self._semaphore.acquire()
+
+    async def __aexit__(self, exc_type, exc_value, traceback):
+        self._semaphore.release()
+
+
 class SessionBase(asyncio.Protocol):
     '''Base class of networking sessions.
 
@@ -89,6 +126,18 @@ class SessionBase(asyncio.Protocol):
     '''
 
     max_errors = 10
+    # Multiply this by bandwidth bytes used to get resource usage cost
+    bw_cost_per_byte = 1 / 100000
+    # If cost is over this requests begin to get delayed and concurrency is reduced
+    cost_soft_limit = 10000
+    # If cost is over this the session is closed
+    cost_hard_limit = 50000
+    # Resource usage is reduced by this every second
+    cost_decay_per_sec = cost_hard_limit / 3600
+    # Request delay ranges from 0 to this between cost_soft_limit and cost_hard_limit
+    cost_sleep = 5.0
+    # Initial number of requests that can be concurrently processed
+    initial_concurrent = 6
 
     def __init__(self, *, framer=None, loop=None):
         self.framer = framer or self.default_framer()
@@ -116,35 +165,14 @@ class SessionBase(asyncio.Protocol):
         self.recv_count = 0
         self.recv_size = 0
         self.last_recv = self.start_time
-        # Bandwidth usage per hour before throttling starts
-        self.bw_limit = 2000000
-        self.bw_time = self.start_time
-        self.bw_charge = 0
+        # Resource usage
+        self.cost = 0.0
+        self._cost_last = 0.0
+        self._cost_time = self.start_time
+        self._cost_fraction = 0.0
         # Concurrency control
-        self.max_concurrent = 6
-        self._concurrency = Concurrency(self.max_concurrent)
-
-    async def _update_concurrency(self):
-        # A non-positive value means not to limit concurrency
-        if self.bw_limit <= 0:
-            return
-        now = time.time()
-        # Reduce the recorded usage in proportion to the elapsed time
-        refund = (now - self.bw_time) * (self.bw_limit / 3600)
-        self.bw_charge = max(0, self.bw_charge - int(refund))
-        self.bw_time = now
-        # Reduce concurrency allocation by 1 for each whole bw_limit used
-        throttle = int(self.bw_charge / self.bw_limit)
-        target = max(1, self.max_concurrent - throttle)
-        current = self._concurrency.max_concurrent
-        if target != current:
-            self.logger.info(f'changing task concurrency from {current} '
-                             f'to {target}')
-            await self._concurrency.set_max_concurrent(target)
-
-    def _using_bandwidth(self, size):
-        '''Called when sending or receiving size bytes.'''
-        self.bw_charge += size
+        self._concurrency = Concurrency(self.initial_concurrent)
+        self._cost_check_delta = max(self.cost_soft_limit // 4, 100)
 
     def _receive_messages(self):
         raise NotImplementedError
@@ -164,7 +192,7 @@ class SessionBase(asyncio.Protocol):
         if not self.is_closing():
             framed_message = self.framer.frame(message)
             self.send_size += len(framed_message)
-            self._using_bandwidth(len(framed_message))
+            self.bump_cost(len(framed_message) * self.bw_cost_per_byte)
             self.send_count += 1
             self.last_send = time.time()
             if self.verbosity >= 4:
@@ -187,7 +215,7 @@ class SessionBase(asyncio.Protocol):
         if self.verbosity >= 4:
             self.logger.debug(f'Received framed message {framed_message}')
         self.recv_size += len(framed_message)
-        self._using_bandwidth(len(framed_message))
+        self.bump_cost(len(framed_message) * self.bw_cost_per_byte)
         self.framer.received_bytes(framed_message)
 
     def pause_writing(self):
@@ -231,6 +259,43 @@ class SessionBase(asyncio.Protocol):
         self._can_send.set()
 
     # External API
+    def bump_cost(self, delta):
+        # Delta can be positive or negative
+        self.cost = max(0, self.cost + delta)
+        if abs(self.cost - self._cost_last) > self._cost_check_delta:
+            self.recalc_concurrency()
+
+    def recalc_concurrency(self):
+        '''Call to recalculate sleeps and concurrency for the session.  Called automatically if
+        cost has drifted significantly.  Otherwise can be called at regular intervals if
+        desired.
+        '''
+        # Refund resource usage proportionally to elapsed time; the bump passed is negative
+        now = time.time()
+        self.cost = max(0, self.cost - (now - self._cost_time) * self.cost_decay_per_sec)
+        self._cost_time = now
+        self._cost_last = self.cost
+
+        # Setting cost_hard_limit <= 0 means to not limit concurrency
+        value = self._concurrency.max_concurrent
+        cost_soft_range = self.cost_hard_limit - self.cost_soft_limit
+        if cost_soft_range <= 0:
+            return
+
+        cost = self.cost + self.extra_cost()
+        self._cost_fraction = max(0.0, (cost - self.cost_soft_limit) / cost_soft_range)
+
+        target = max(0, ceil((1.0 - self._cost_fraction) * self.initial_concurrent))
+        if target != value:
+            self.logger.info(f'changing task concurrency from {value} to {target}')
+        self._concurrency.set_target(target)
+
+    def extra_cost(self):
+        '''A dynamic value added to this session's cost when deciding how much to throttle
+        requests.  Can be negative.
+        '''
+        return 0.0
+
     def default_framer(self):
         '''Return a default framer.'''
         raise NotImplementedError
@@ -311,23 +376,25 @@ class MessageSession(SessionBase):
             else:
                 self.last_recv = time.time()
                 self.recv_count += 1
-                if self.recv_count % 10 == 0:
-                    await self._update_concurrency()
                 await self._throttled_message(message)
 
     async def _throttled_message(self, message):
         '''Process a single request, respecting the concurrency limit.'''
-        async with self._concurrency.semaphore:
-            try:
+        try:
+            async with self._concurrency:
+                if self._cost_fraction:
+                    await sleep(self._cost_fraction * self.cost_sleep)
                 await self.handle_message(message)
-            except ProtocolError as e:
-                self.logger.error(f'{e}')
-                self._bump_errors()
-            except CancelledError:
-                raise
-            except Exception:
-                self.logger.exception(f'exception handling {message}')
-                self._bump_errors()
+        except ProtocolError as e:
+            self.logger.error(f'{e}')
+            self._bump_errors()
+        except ExcessiveSessionCostError:
+            self._close()
+        except CancelledError:
+            raise
+        except Exception:
+            self.logger.exception(f'exception handling {message}')
+            self._bump_errors()
 
     # External API
     def default_framer(self):
@@ -435,8 +502,6 @@ class RPCSession(SessionBase):
 
             self.last_recv = time.time()
             self.recv_count += 1
-            if self.recv_count % 10 == 0:
-                await self._update_concurrency()
 
             try:
                 requests = self.connection.receive_message(message)
@@ -453,26 +518,31 @@ class RPCSession(SessionBase):
 
     async def _throttled_request(self, request):
         '''Process a single request, respecting the concurrency limit.'''
-        async with self._concurrency.semaphore:
-            try:
+        try:
+            async with self._concurrency:
+                if self._cost_fraction:
+                    await sleep(self._cost_fraction * self.cost_sleep)
                 result = await self.handle_request(request)
-            except (ProtocolError, RPCError) as e:
-                result = e
-            except CancelledError:
-                raise
-            except Exception:
-                self.logger.exception(f'exception handling {request}')
-                result = RPCError(JSONRPC.INTERNAL_ERROR,
-                                  'internal server error')
-            if isinstance(request, Request):
-                message = request.send_result(result)
-                if message:
-                    await self._send_message(message)
+        except (ProtocolError, RPCError) as e:
+            result = e
+        except ExcessiveSessionCostError:
+            result = FinalRPCError(JSONRPC.EXCESSIVE_RESOURCE_USAGE, 'excessive resource usage')
+        except CancelledError:
+            raise
+        except Exception:
+            self.logger.exception(f'exception handling {request}')
+            result = RPCError(JSONRPC.INTERNAL_ERROR,
+                              'internal server error')
+
+        if isinstance(request, Request):
+            message = request.send_result(result)
+            if message:
+                await self._send_message(message)
+        if isinstance(result, Exception):
+            self._bump_errors()
             if isinstance(result, FinalRPCError):
                 # Don't await self.close() because that is self-cancelling
                 self._close()
-            elif isinstance(result, Exception):
-                self._bump_errors()
 
     def connection_lost(self, exc):
         # Cancel pending requests and message processing
